/*
Version 22.03
=============
*/

const { Clutter, Gio, GObject, Meta, Shell, St, Gtk } = imports.gi;

const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Config = imports.misc.config;
const ExtensionUtils = imports.misc.extensionUtils;

const Me = ExtensionUtils.getCurrentExtension();
const NoticeApplet = Me.imports.noticeApplet;
const EasySupport = Me.imports.easySupport;
const NotificationSection = Me.imports.notificationSection;

const MetaKeyBindingFlags = Meta.KeyBindingFlags;

const ShellActionMode = Shell.ActionMode;

const Gettext = imports.gettext;
const _ = Gettext.domain('notification-center').gettext;

const MAX_NOTIFICATION_TEXT = ' 99+ ';
const DEFAULT_X_MARGIN = 22;
const DEFAULT_Y_MARGIN = 7;
const DIGIT_X_MARGIN = 6;

let notificationCenter = null;
let originalNotificationSection = null;
var useNoticeButton = false;

function init() { // eslint-disable-line no-unused-vars
    Gettext.bindtextdomain('notification-center', Config.LOCALEDIR);
}

function enable() {
    notificationCenter = new NotificationCenter();
    notificationCenter._enable();
    reloadExtensionOnPrefsChange();
    reloadApplicationProfilesOnPrefsChange();
}

function disable() {
    notificationCenter._onDestroy();
    notificationCenter.destroy();
}

function reloadApplicationProfilesOnPrefsChange() {
    // Reloads Application Profiles when preferences are changed.
    notificationCenter.reloadProfilesSignal = notificationCenter.prefs.connect('changed::reload-profiles-signal', () => notificationCenter.loadPreferences());
}

function reloadExtensionOnPrefsChange() {
    // Reloads the Extension when preferences are changed.
    notificationCenter.reloadSignal = notificationCenter.prefs.connect('changed::reload-signal', () => {
        disable();
        enable();
    });
}

var NotificationCenter = GObject.registerClass(class NotificationCenter extends PanelMenu.Button {
    _init() {
        this.prefs = ExtensionUtils.getSettings('org.gnome.shell.extensions.notification-center');
        super._init(0.5, null, true);
        this.replacePopupMenu();

        this._messageList = Main.panel.statusArea.dateMenu._messageList;
        this.mediaSection = this._messageList._mediaSection;
        this.notificationSection = new NotificationSection.NotificationSectionByGroup();
        this.eventsSection = Main.panel.statusArea.dateMenu._eventsItem;
        this._messageListParent = this._messageList.get_parent();
        this.newEventsSectionParent = this.eventsSection.get_parent();

        this.loadPreferences();
        this.connectedSignals = [];
        this.dmsig = null;
        this.cmsig = null;
        this.dndSig = null;
        this.reloadSignal = null;
        this.reloadProfilesSignal = null;

        this.textureCache = St.TextureCache.get_default();
        this.iconThemeChangeSig = null;
        this.notificationIconName = null;

        this.eventsCount = 0;
        this.mediaCount = 0;
        this.seenEvents = false;
        this.messageListRemoved = false;
        this.isDndOff = true;
        this.dndpref = Main.panel.statusArea.dateMenu._indicator._settings;

        this.initLayout();
    }


    replacePopupMenu() {
        // replace builtin popupmenu into new one
        let menuManager = new PopupMenu.PopupMenuManager(Main.panel);
        let section = new PopupMenu.PopupMenuSection();
        this.newMenu = new PopupMenu.PopupMenu(this, 0.5, St.Side.TOP);
        this.notificationCenterBox = new St.BoxLayout({
            vertical: true,
            x_expand: true,
            y_expand: true,
        });

        this.mainBox = new St.BoxLayout({
            vertical: true,
            x_expand: true,
            y_expand: true,
        });

        this.menu.destroy();
        Main.uiGroup.add_actor(this.newMenu.actor);
        this.newMenu.actor.hide();
        this.mainBox._delegate = this.mainBox;
        this.newMenu.connect('open-state-changed', this._onOpenStateChanged.bind(this));

        section.actor.add_actor(this.notificationCenterBox);
        this.notificationCenterBox.add_child(this.mainBox);
        this.newMenu.addMenuItem(section);
        menuManager.addMenu(this.newMenu);

        this.newMenu.actor.remove_style_class_name('popup-menu-boxpointer');
        this.newMenu.actor.add_style_class_name('notification-center-actor');
        this.newMenu.box.style = 'margin:0px; padding: 0px;';
        this.notificationCenterBox.add_style_class_name('notification-center-popupmenu-total');
        this.mainBox.add_style_class_name('notification-center-popupmenu');
        // bug?
        // To show shadow of this.mainBox,this.newMenu should have css properfies
        // -arrow-border-width:12px and -arrow-rise:-12px
        // maybe cairo drawing bug, because St.drawing in Popupmenu is using cairo
    }

    vfunc_event(event) {
        if (event.type() === Clutter.EventType.BUTTON_PRESS) {
            if (event.get_button() === 1)
                this.toggleMenu();

        } else if (event.type() === Clutter.EventType.TOUCH_BEGIN) {
            this.toggleMenu();
        }
        return Clutter.EVENT_PROPAGATE;
    }

    toggleMenu() {
        this.newMenu.toggle();
        if (this.newMenu.isOpen)
            this.mainBox.grab_key_focus();
    }

    _enable() {
        this.add_child(this._indicator);

        this._indicator.add_child(this.eventsIcon);
        this._indicator.add_child(this.eventsLabel);
        this._indicator.add_child(this.mediaIcon);
        this._indicator.add_child(this.notificationIcon);
        this._indicator.add_child(this._labelBin);

        this.mainBox.add_actor(this.titleBox);

        originalNotificationSection = this._messageList._notificationSection;

        // update notification icon when theme changed
        // this.iconThemeChangeSig = this.textureCache.connect('icon-theme-changed', this.iconThemeChanged.bind(this));

        Main.panel.addToStatusArea('NotificationCenter', this, this.prefs.get_int('indicator-index'), this.prefs.get_string('indicator-pos'));
        // this.rebuildMessageList();
        this._messageListParent.remove_actor(this._messageList);
        this._messageListParent.insert_child_at_index(this._messageList, this.messageListPos);

        for (let i = 0; i < this.showingSections.length; i++) {
            if (this.showingSections[i] === 'events') {
                this.newEventsSectionParent.remove_actor(this.eventsSection);
                this.box.add(this.eventsSection);
                this.connectedSignals.push(this.eventsSection._eventsList.connect('actor-added', () => this.newNotif(this.showingSections[i])));
                this.connectedSignals.push(this.eventsSection._eventsList.connect('actor-removed', () => this.remNotif(this.showingSections[i])));
                this.eventsSection.setDate(new Date());
                this[`${this.showingSections[i]}Section`].add_style_class_name('notification-center-message-list-section');
            } else if (this.showingSections[i] === 'notification') {
                this._removeSection(originalNotificationSection);
                this.box.add(this.notificationSection);
                this.connectedSignals.push(this.notificationSection.connect('noti-added', () => this.newNotif('notification')));
                this.connectedSignals.push(this.notificationSection.connect('noti-destroyed', () => this.newNotif('notification')));

            } else {
                this._removeSection(this[`${this.showingSections[i]}Section`]);
                this.box.add(this[`${this.showingSections[i]}Section`]);
                this.connectedSignals.push(this[`${this.showingSections[i]}Section`]._list.connect('actor-added', () => this.newNotif(this.showingSections[i])));
                this.connectedSignals.push(this[`${this.showingSections[i]}Section`]._list.connect('actor-removed', () => this.remNotif(this.showingSections[i])));
                this[`${this.showingSections[i]}Section`].add_style_class_name('notification-center-message-list-section');
            }
        }
        Main.messageTray._bannerBin.connect('actor-added', () => this._onNewBannerAdded());
        // this.arrangeItems();
        this.scrollView._delegate = this;
        this.scrollView.add_actor(this.box);
        this.mainBox.add_child(this.scrollView);
        // this.addClearButton()
        if (this.prefs.get_enum('clear-button-alignment') === 3)
            return;

        this.clearButton.connect('clicked', () => {
            this.notificationSection.clear();
        });
        //        this.clearButton.set_x_align(1 + this.prefs.get_enum('clear-button-alignment'));
        this.mainBox.add_child(this.emptyIcon);
        this.mainBox.add_child(this.emptyLabel);
        this.titleBox.add_actor(this.titleLabel);
        this.titleBox.add_actor(this.clearButton);

        if (useNoticeButton)
            this.mainBox.add_actor(this.noticeButton);


        // add easysupport section
        this.notificationCenterBox.add_actor(this.easySupport);
        this.easySupport.makeSection();
        this.easySupport.alarmIcon._button.connect('clicked', () => this.dndToggle());

        this.loadDndStatus();
        this.resetIndicator();

        Main.messageTray.bannerAlignment = this.prefs.get_enum('banner-pos');
        Main.messageTray.set_y_align(Clutter.ActorAlign.END);
        // this.removeDotAndBorderFromDateMenu();
        Main.panel.statusArea.dateMenu.get_children()[0].remove_actor(Main.panel.statusArea.dateMenu._indicator);
        this.dtActors = Main.panel.statusArea.dateMenu.get_children()[0].get_children();
        Main.panel.statusArea.dateMenu.get_children()[0].remove_actor(this.dtActors[0]);

        if (this.showingSections.length === 3 && !this.showEventsInCalendarAlso)
            this._messageListParent.get_children()[1].style = 'RIGHTer-width: 0px';

        // this.indicatorViewShortcut();
        Main.wm.addKeybinding(
            'indicator-shortcut',
            this.prefs,
            MetaKeyBindingFlags.NONE,
            ShellActionMode.NORMAL | ShellActionMode.OVERVIEW | ShellActionMode.POPUP,
            () => {
                this.notificationIcon.visible = !(this.mediaIcon.visible || this.eventsIcon.visible);
                this.visible = true;
                this.newMenu.toggle();
            }
        );

        this.dndSig = this.dndpref.connect('changed::show-banners', () => {
            this.loadDndStatus();
        });

        if (this.prefs.get_boolean('middle-click-dnd'))
            this.connect('button-press-event', (actor, event) => this.middleClickDndToggle(actor, event));


        this.dmSig = Main.panel.statusArea.dateMenu.menu.connect('open-state-changed', () => {
            if (Main.panel.statusArea.dateMenu.menu.isOpen) {
                switch (this.hideEmptySpace) {
                case 0:
                    this.manageEvents(1);
                    if (this.showLabel === false)
                        this.eventsCount = 0;

                    this.resetIndicator();
                    break;

                case 2:
                    if (!this.mediaSectionToBeShown && this.mediaSection._shouldShow() || !this.notificationSectionToBeShown && !this.notificationSection.isEmpty() || !this.eventsSectionToBeShown &&  this.shouldShowEventsSection()) {
                        if (this.messageListRemoved) {
                            this._messageListParent.insert_child_at_index(this._messageList, this.messageListPos);
                            this.messageListRemoved = false;
                        }
                    } else if (!this.messageListRemoved) {
                        this._messageListParent.remove_actor(this._messageList);
                        this.messageListRemoved = true;
                    }
                    break;
                }
            } else {
                Main.panel.statusArea.dateMenu._calendar.setDate(new Date());
                this.eventsCount = this.seenEvents ? 0 : this.eventsCount;
                this.resetIndicator();
            }
        });

        if (this.menuAutoclose) {
            this.cmsig = global.display.connect('notify::focus-window', () => {
                if (global.display.focus_window !== null && this.newMenu.isOpen)
                    this.newMenu.close(1);
            });
        }

        this.defaultWeatherItemVisibility = Main.panel.statusArea.dateMenu._weatherItem.visible;
        Main.panel.statusArea.dateMenu._weatherItem.visible = !this.prefs.get_boolean('hide-weather-section') && this.defaultWeatherItemVisibility;
        this.defaultClocksItemVisibility = Main.panel.statusArea.dateMenu._clocksItem.visible;
        Main.panel.statusArea.dateMenu._clocksItem.visible =  !this.prefs.get_boolean('hide-clock-section') && this.defaultClocksItemVisibility;
        this.defaultDateVisibility = Main.panel.statusArea.dateMenu._date.visible;
        Main.panel.statusArea.dateMenu._date.visible =  !this.prefs.get_boolean('hide-date-section') && this.defaultDateVisibility;

        this.notificationSection._messages.forEach(message => this._setMessageLayout(message));

        this.subPanelElements = [];
        this.changeSubPanelUI();
        this.changeSubPanelUILater();
        if (this.noticeApplet)
            this.noticeApplet.enable();
    }

    changeSubPanelUI() {
        // apply UI changes also to sub panels
        if (!global.dashToPanel)
            return false;
        if (!global.dashToPanel.panels)
            return false;
        if (global.dashToPanel.panels.length < 2)
            return false;

        global.dashToPanel.panels.forEach(panel => {
            if (panel === Main.panel.get_parent())
                return;

            // remove messagelist, dots and borders of datemenu in sub panel
            // each elements saved as [messageList, messageListParent, dtActor, dtActorParents, indicators]
            let subPanelElement = [];
            let messageList = panel.statusArea.dateMenu._messageList;
            let messageListParent = messageList.get_parent();
            if (!messageListParent)
                return;

            let dtActor = panel.statusArea.dateMenu.get_children()[0].get_children()[0];
            let dtActorParent = panel.statusArea.dateMenu.get_children()[0];
            let indicator = panel.statusArea.dateMenu._indicator;

            subPanelElement.push(messageList);
            subPanelElement.push(messageListParent);
            subPanelElement.push(dtActor);
            subPanelElement.push(dtActorParent);
            subPanelElement.push(indicator);

            messageListParent.remove_actor(messageList);
            dtActorParent.remove_actor(dtActor);
            dtActorParent.remove_actor(indicator);

            this.subPanelElements.push(subPanelElement);

            // use same default visitility value of main panel
            panel.statusArea.dateMenu._weatherItem.visible = !this.prefs.get_boolean('hide-weather-section') && this.defaultWeatherItemVisibility;
            panel.statusArea.dateMenu._clocksItem.visible =  !this.prefs.get_boolean('hide-clock-section') && this.defaultClocksItemVisibility;
            panel.statusArea.dateMenu._date.visible =  !this.prefs.get_boolean('hide-date-section') && this.defaultDateVisibility;
        });

        return true;
    }

    changeSubPanelUILater() {
        // register callback function for later load of dash-to-panel
        this.subPanelId =  Main.extensionManager.connect('extension-state-changed', (data, extension) => {
            if (extension.uuid === 'dash-to-panel@tmax-shell-extensions' && extension.state === 1)
                this.changeSubPanelUI();
        });

        this._monitorsChangedId =
            Main.layoutManager.connect('monitors-changed', this.changeSubPanelUI.bind(this));
    }

    initLayout() {
        this.eventsIcon = new St.Icon({ style_class: 'system-status-icon',
            visible: false,
            icon_name: 'x-office-calendar-symbolic' });
        this.mediaIcon = new St.Icon({ style_class: 'system-status-icon',
            visible: false,
            icon_name: 'audio-x-generic-symbolic' });
        this.notificationIcon = new St.Icon({ style_class: 'system-status-icon', visible: false,
            gicon: Gio.icon_new_for_string(`${Me.path}/img/tos_ic_notification.svg`),
            icon_size: 24 });
        this.emptyIcon = new St.Icon({ style_class: 'notification-center-empty-icon',
            gicon: Gio.icon_new_for_string(`${Me.path}/img/tos_img_alarm_none.svg`),
            icon_size: 150,
            x_align: Clutter.ActorAlign.CENTER,
            visible: false });

        this.emptyLabel = new St.Label({ style_class: 'notification-center-empty-label',
            text: _('No New Notifications'),
            x_align: Clutter.ActorAlign.CENTER,
            visible: false });
        this.titleLabel = new St.Label({ style_class: 'notification-center-title-label',
            text: _('Notification Center') });
        this.eventsLabel = new St.Label({ text: '• ',
            visible: false });

        this._initNotificationLabel();

        this._indicator = new St.BoxLayout({ style_class: 'panel-status-menu-box',
            style: 'spacing:0.0em' });
        this._indicator.add_style_class_name('panel-indicator');
        this.box = new St.BoxLayout({ style_class: 'notification-center-message-list',
            vertical: true });
        this.titleBox = new St.BoxLayout({ style_class: 'notification-center-title-box',
            vertical: false,
            y_align: Clutter.ActorAlign.START,
            x_expand: true });

        this.clearButton = new St.Button({ style_class: 'notification-center-clear-button',
            label: _('Clear'),
            can_focus: true,
            visible: false,
            x_expand: true });
        this.easySupport = new EasySupport.EasySupport();

        this.scrollView = new St.ScrollView({
            x_expand: false,
            y_expand: false,
            x_align: Clutter.ActorAlign.START,
            y_align: Clutter.ActorAlign.START,
        });
        this.scrollView.set_policy(St.PolicyType.NEVER, St.PolicyType.AUTOMATIC);
        this.scrollView.clip_to_allocation = true;


        this.add_style_class_name('notification-center-panel-button');
        this.notificationIcon.set_pivot_point(0.5, 0);

        if (useNoticeButton) {
            this.noticeButton = new St.Button({ style_class: 'notification-center-clear-button button',
                label: _('Notice'),
                can_focus: true,
                x_align: Clutter.ActorAlign.START,
                y_align: Clutter.ActorAlign.END,
                x_expand: true });
            this.noticeButton.set_toggle_mode(true);

            this.noticeApplet = new NoticeApplet.noticeApplet();
            this.noticeApplet.init(this);
        }
    }

    _initNotificationLabel() {
        this._labelBin = new St.Widget();

        this._labelPositionPath = new Clutter.Path();
        this._labelPositionPath.add_move_to(DEFAULT_X_MARGIN, DEFAULT_Y_MARGIN);
        let labelConstraint = new Clutter.PathConstraint({
            offset: 0,
            path: this._labelPositionPath,
        });

        this._labelBin.add_constraint(labelConstraint);

        this.notificationLabel = new St.Label({ y_expand: false,
            x_expand: false,
            visible: false });
        this._labelBin.add_child(this.notificationLabel);
    }

    animateOnNewNotification(times, op = 254, angle = 3) {
        if (times === 0 || !this.animateIcon) {
            this.notificationIcon.ease({
                duration: 150,
                scale_x: 1.0,
                scale_y: 1.0,
                translation_y: 0,
                opacity: 255,
                rotation_angle_z: 0,
                onComplete: () =>  this.blinkIcon(!this.newMenu.isOpen * this.blinkCount, this.blinkTime, 100),
            });
            return;
        }

        [this.visible, this.notificationIcon.visible] = [true, true];

        this.notificationIcon.ease({
            duration: 150,
            scale_x: 1.2,
            scale_y: 1.2,
            translation_y: -4,
            opacity: op,
            rotation_angle_z: angle,
            onComplete: () =>  this.animateOnNewNotification(--times, op - 1, -angle),
        });
    }

    blinkIcon(blinkTimes, interval, opacity) {
        if (blinkTimes > 0) {
            this.notificationIcon.ease({
                duration: interval,
                opacity,
                onComplete: () => this.blinkIcon(--blinkTimes, interval, opacity === 255 ? 100 : 255),
            });
        }
    }

    blinkIconStopIfBlinking() {
        this.notificationIcon.remove_all_transitions();
        this.notificationIcon.set_opacity(255);
    }

    dndToggle() {
        this.dndpref.set_boolean('show-banners', !this.dndpref.get_boolean('show-banners'));
    }

    /*
  loadDndStatus is legacy code, but we need to use this function, then renaming later.
  it means that rename from loadDndStatus to loadAlarmStatus.

  If show-banner is true, then Do not disturb function is disabled
  So, If show banners is true, than 'Do not disturb' is off
  Above reason, isDndoff = get_value(show-banners)
*/
    loadDndStatus() {
        this.isDndOff = this.dndpref.get_boolean('show-banners');

        this.blinkIconStopIfBlinking();
        this.manageAutohide();

        if (this.notificationIconName)
            this.notificationIcon.icon_name = this.notificationIconName;

        else
            this.notificationIcon.gicon = Gio.icon_new_for_string(`${Me.path}/img/tos_ic_notification.svg`);

        if (this.isDndOff) {
            this.easySupport.alarmIcon.setStatus(true);
            this.notificationIcon.set_opacity(255);
            this.manageLabel();
            return false;
        }

        this.notificationIcon.gicon = Gio.icon_new_for_string(`${Me.path}/img/tos_ic_notification_no.svg`);

        this.easySupport.alarmIcon.setStatus(false);
        Main.messageTray._bannerBin.hide();
        this.notificationLabel.hide();
        this.eventsLabel.hide();
        return true;
    }

    loadPreferences() {
        this.autohide                     = this.prefs.get_int('autohide');
        this.mediaSectionToBeShown        = this.prefs.get_int('show-media') > 0;
        this.notificationSectionToBeShown = this.prefs.get_int('show-notification') > 0;
        this.eventsSectionToBeShown       = this.prefs.get_int('show-events') > 0;
        this.hideEmptySpace               = this.prefs.get_enum('beside-calendar');
        this.showEventsInCalendarAlso     = this.eventsSectionToBeShown ? this.hideEmptySpace === 0 : false;
        this.hideEventsSectionIfEmpty     = !this.prefs.get_boolean('hide-events-section-if-empty');
        this.showThreeIcons               = this.prefs.get_boolean('individual-icons');
        this.includeEventsCount           = this.prefs.get_boolean('include-events-count');
        this.newNotificationAction        = this.prefs.get_enum('new-notification');
        this.menuAutoclose                = this.prefs.get_boolean('autoclose-menu');
        this.eventsSectionhere            = this.showEventsInCalendarAlso;
        this.showingSections              = this.prefs.get_strv('sections-order');
        this.messageListPos               = this.prefs.get_boolean('calendar-on-left') ? 1 : 0;
        this.appBlackList                 = this.prefs.get_strv('name-list');
        this.blackListAction              = this.prefs.get_enum('for-list');
        this.animateIcon                  = this.prefs.get_boolean('animate-icon');
        this.blinkTime                    = this.prefs.get_int('blink-time');
        this.blinkCount                   = this.prefs.get_int('blink-icon') * 2;
        this.showLabel                    = this.prefs.get_boolean('show-label');
        this.changeIcons                  = this.prefs.get_boolean('change-icons');
    }

    manageAutohide() {
        if (!this.newMenu.isOpen) {
            this.mediaIcon.visible        = this.mediaSection._shouldShow() && this.showThreeIcons && this.mediaSectionToBeShown;
            this.eventsIcon.visible       = this.shouldShowEventsSection() && this.showThreeIcons && this.eventsSectionToBeShown;
            this.notificationIcon.visible = !this.notificationSection.isEmpty() && this.notificationSectionToBeShown ||
                                      this.mediaSection._shouldShow() && this.mediaSectionToBeShown && !this.showThreeIcons ||
                                      this.shouldShowEventsSection() && this.eventsSectionToBeShown && !this.showThreeIcons ||
                                      !this.isDndOff * this.autohide > 1;
            if (this.mediaIcon.visible || this.eventsIcon.visible || this.notificationIcon.visible || !this.autohide) {
                this.visible = true;
                this.notificationIcon.visible = this.mediaIcon.visible || this.eventsIcon.visible ? this.notificationIcon.visible : true;
                return;
            }
            this.visible = false;
        }
    }

    manageEvents(action) {
        this.eventsSection.visible = this.shouldShowEventsSection() || this.hideEventsSectionIfEmpty;

        if (this.showEventsInCalendarAlso === true) {
            switch (action) {
            case 0:
                if (this.eventsSectionhere === true)
                    return;

                this._removeSection(this.eventsSection);
                this.box.insert_child_at_index(this.eventsSection, this.showingSections.indexOf('events'));
                this.eventsSectionhere = true;
                return;
            case 1:
                if (this.eventsSectionhere === false)
                    return;

                this.box.remove_child(this.box.get_children()[this.showingSections.indexOf('events')]);
                this.newEventsSectionParent.insert_child_at_index(this.eventsSection, 0);
                this.eventsSectionhere = false;
            }
        }
    }

    manageLabel(nCount, eCount) {
        this.notificationLabel.visible = nCount * this.newNotificationAction;
        this.eventsLabel.visible = eCount * this.newNotificationAction && this.shouldShowEventsSection() > 0;

        if (this.changeIcons)
            this.manageIconChange(nCount > 0 || eCount > 0);

        if (this.newNotificationAction === 2) {

            this._labelPositionPath.clear();
            if (nCount > 0) {
                if (nCount > 99) {
                    this._labelPositionPath.add_move_to(DEFAULT_X_MARGIN - 2 * DIGIT_X_MARGIN, DEFAULT_Y_MARGIN);
                    this.notificationLabel.text = MAX_NOTIFICATION_TEXT;
                } else if (nCount > 9) {
                    this._labelPositionPath.add_move_to(DEFAULT_X_MARGIN - DIGIT_X_MARGIN, DEFAULT_Y_MARGIN);
                    this.notificationLabel.text = ` ${nCount.toString()} `;
                } else {
                    this._labelPositionPath.add_move_to(DEFAULT_X_MARGIN, DEFAULT_Y_MARGIN);
                    this.notificationLabel.text = ` ${nCount.toString()} `;
                }
                if (this.notificationSection.isUrgent())
                    this.notificationLabel.set_style_class_name('notification-center-noti-label-urgent');
                else
                    this.notificationLabel.set_style_class_name('notification-center-noti-label');
            }
            if (eCount > 0)
                this.eventsLabel.text = `${eCount.toString()} `;
        }
    }

    manageIconChange(statusIcon) {
        let iconName = statusIcon ? 'notification-center-full' : 'notification-center-empty';
        this.notificationIcon.icon_name = iconName;
    }

    middleClickDndToggle(actor, event) {
        switch (event.get_button()) {
        case 2: // if middle click
            // close the menu, since it gets open on any click
            if (this.newMenu.isOpen)
                this.newMenu.actor.hide();

            // toggle DND state
            this.dndToggle();
            // reload dnd status
            this.loadDndStatus();
        }
    }

    newNotif(messageType) {
        Main.messageTray._bannerBin.visible = true;
        switch (messageType) {
        case 'media':
            this.mediaCount++;
            break;
        case 'events':
            [this.seenEvents, this.eventsCount] = [Main.panel.statusArea.dateMenu.menu.isOpen ? this.seenEvents : false, this.eventsCount + !this.newMenu.isOpen];
            break;
        }
        this.resetIndicator();
    }

    _onNewBannerAdded() {
        if (!Main.messageTray._banner) {
            global.log('There is no banner in messageTary');
            return;
        }
        Main.messageTray._banner.add_style_class_name('notification-center-banner');
    }

    remNotif(messageType) {
        switch (messageType) {
        case 'media':
            this.mediaCount--;
            break;
        case 'events':
            if (this.eventsCount > 0)
                this.eventCount--;
            break;
        }
        this.resetIndicator();
    }

    _removeSection(section) {
        if (section === this.eventsSection) {
            this.newEventsSectionParent.remove_actor(this.eventsSection);
            return;
        }

        this._messageList._sectionList.remove_actor(section);
        this._messageList._sync();
    }

    resetIndicator() {
        this.manageAutohide();
        this.clearButton.visible = this.notificationSection.notiCount !== 0 && this.notificationSectionToBeShown;
        this.emptyIcon.visible = !this.clearButton.visible;
        this.emptyLabel.visible = !this.clearButton.visible;
        this.eventsCount = this.eventsCount * this.includeEventsCount;

        if (this.isDndOff)
            this.manageLabel(this.notificationSection.notiCount + !this.showThreeIcons * this.eventsCount, this.showThreeIcons * this.eventsCount);
    }

    _onOpenStateChanged(menu, open) {
        if (!open) {
            this.resetIndicator();
            this.remove_style_pseudo_class('active');
            return;
        }

        this.add_style_pseudo_class('active');
        this.manageEvents(0);
        [this.mediaSection.visible, this.notificationSection.visible] = [true, true];
        this.blinkIconStopIfBlinking();

        if (!this.showLabel)
            this.eventsCount = 0;

        this.seenEvents = true;
        this.resetIndicator();

        let dashToPanelSetting = new Gio.Settings({ schema_id: 'org.gnome.shell.extensions.dash-to-panel' });
        let dashToPanelSize = dashToPanelSetting.get_int('panel-size');

        // TODO(hyeongmin, 230303): to prevent style override, get previous style and add max-height.
        // Set 1px more margin to newMenu max-height. Without it, GUI breaks.
        this.newMenu.actor.style = `max-height : ${Main.layoutManager.primaryMonitor.height - dashToPanelSize - 1}px`;
    }

    setNotificationIconName() {
        this.notificationIconName = null;
        if (Gtk.IconTheme.get_default())
            this.notificationIconName = Gtk.IconTheme.get_default().has_icon('notification-symbolic') ? 'notification-symbolic' : 'preferences-system-notifications-symbolic';
        else
            this.notificationIconName = 'preferences-system-notifications-symbolic';
    }

    iconThemeChanged() {
        this.setNotificationIconName();
        this.loadDndStatus();
    }

    shouldShowEventsSection() {
        switch (this.eventsSection._eventsList.get_children().length) {
        case 0:
            return 0;
        default:
            return this.eventsSection._eventsList.get_children()[0].text === _('No Events') ? 0 : this.eventsSection._eventsList.get_children().length;
        }
    }

    _setMessageLayout(message)  {
        // check whether message's timelabel moved or not
        // timelabel move should be done only once
        if (!message._secondaryBin.child)
            return;
        message.add_style_class_name('notification-center-message');
        message._bodyStack.style_class = 'notification-center-bodystack';
        // move timelabel from side of title to down of body
        let contentBox = message.child.get_child_at_index(0).get_child_at_index(1);
        let timelabel = message._secondaryBin.child;
        message._secondaryBin.child = null;
        contentBox.add_child(timelabel);
        message.setExpandedLines(3);
        message.expand();
    }

    revertSubPanelUI() {
        // revert UI changes in sub panel
        if (!global.dashToPanel)
            return;
        if (!global.dashToPanel.panels)
            return;
        if (global.dashToPanel.panels.length < 2)
            return;

        global.dashToPanel.panels.forEach(panel => {
            if (panel === Main.panel.get_parent())
                return;

            // revert messagelist, dots and borders of datemenu in sub panel
            // uses elements saved as [messageList, messageListParent, dtActor, dtActorParents, indicators] in changeSubPanel()
            // subPanelElement[0] => original messagelist in sub panel
            // subPanelElement[1] => parent of original messagelist
            // subPanelElement[2] => border of datemenu
            // subPanelElement[3] => datemenu actor
            // subPanelElement[4] => dot of datemenu, indicator
            let subPanelElement = this.subPanelElements.shift();
            subPanelElement[1].insert_child_at_index(subPanelElement[0], 0);
            subPanelElement[3].insert_child_at_index(subPanelElement[2], 0);
            subPanelElement[3].add_actor(subPanelElement[4]);

            // revert visibility of items
            panel.statusArea.dateMenu._weatherItem.visible = this.defaultWeatherItemVisibility;
            panel.statusArea.dateMenu._clocksItem.visible = this.defaultClocksItemVisibility;
            panel.statusArea.dateMenu._date.visible = this.defaultDateVisibility;
        });
    }

    _onDestroy() {
        if (this.noticeApplet)
            this.noticeApplet.onDestroy();

        Main.extensionManager.disconnect(this.subPanelId);
        Main.layoutManager.disconnect(this._monitorsChangedId);
        this.revertSubPanelUI();
        this.blinkIconStopIfBlinking();

        if (this.messageListRemoved) {
            this._messageListParent.insert_child_at_index(this._messageList, 0);
            this.messageListRemoved = false;
        } else {
            this._messageListParent.remove_actor(this._messageList);
            this._messageListParent.insert_child_at_index(this._messageList, 0);
        }

        this._messageListParent.get_children()[1].style = '';

        this._messageList._dndSwitch.show();
        this._messageList._dndButton.label_actor.show();

        this.manageEvents(0);
        // this.removeAndDisconnectSections();
        let len = this.showingSections.length;
        while (len !== 0) {

            if (this.showingSections[len - 1] === 'events') {

                this[`${this.showingSections[len - 1]}Section`]._eventsList.disconnect(this.connectedSignals[2 * len - 1]);
                this[`${this.showingSections[len - 1]}Section`]._eventsList.disconnect(this.connectedSignals[2 * len - 2]);

                this.box.remove_child(this.box.get_children()[len - 1]);
                this.newEventsSectionParent.add_actor(this.eventsSection);
            } else {
                this[`${this.showingSections[len - 1]}Section`]._list.disconnect(this.connectedSignals[2 * len - 1]);
                this[`${this.showingSections[len - 1]}Section`]._list.disconnect(this.connectedSignals[2 * len - 2]);

                this.box.remove_child(this.box.get_children()[len - 1]);
                if (this.showingSections[len - 1] === 'notification')
                    this._messageList._addSection(originalNotificationSection);
                else
                    this._messageList._addSection(this[`${this.showingSections[len - 1]}Section`]);
            }
            this[`${this.showingSections[len - 1]}Section`].remove_style_class_name('notification-center-message-list-section');
            this.connectedSignals.pop();
            this.connectedSignals.pop();

            len--;
        }

        this._removeSection(this.mediaSection);
        this._removeSection(this.eventsSection);

        this._messageList._addSection(this.mediaSection);
        this._messageList._addSection(originalNotificationSection);
        this.newEventsSectionParent.insert_child_at_index(this.eventsSection, 0);

        Main.messageTray._bannerBin.show();
        Main.messageTray.bannerAlignment = 2;

        Main.panel.statusArea.dateMenu.menu.disconnect(this.dmSig);

        if (this.menuAutoclose)
            global.display.disconnect(this.cmsig);

        if (this.dndSig !== null)
            this.dndpref.disconnect(this.dndSig);

        if (this.iconThemeChangeSig)
            this.textureCache.disconnect(this.iconThemeChangeSig);

        Main.panel.statusArea.dateMenu.get_children()[0].insert_child_at_index(this.dtActors[0], 0);
        Main.panel.statusArea.dateMenu.get_children()[0].add_actor(Main.panel.statusArea.dateMenu._indicator);

        Main.panel.statusArea.dateMenu._weatherItem.visible = this.defaultWeatherItemVisibility;
        Main.panel.statusArea.dateMenu._clocksItem.visible  = this.defaultClocksItemVisibility;
        Main.panel.statusArea.dateMenu._date.visible        = this.defaultDateVisibility;

        Main.wm.removeKeybinding('indicator-shortcut');

        this.eventsIcon.destroy();
        this.eventsLabel.destroy();
        this.mediaIcon.destroy();
        this.notificationIcon.destroy();
        this.notificationLabel.destroy();
        this._indicator.destroy();

        this.clearButton.destroy();
        this.box.destroy();
        this.scrollView.destroy();
        this.emptyIcon.destroy();
        this.emptyLabel.destroy();
        this.titleLabel.destroy();
        this.easySupport.destroySection();
        this.titleBox.destroy();

        this.prefs.disconnect(this.reloadSignal);
        this.prefs.disconnect(this.reloadProfilesSignal);
    }
});
