const {  GObject, St, Clutter, Gio } = imports.gi;

const MessageList = imports.ui.messageList;
const MessageTray = imports.ui.messageTray;
const Calendar = imports.ui.calendar;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();

const ICON_SIZE = 20;
const NOTI_ICON_SIZE = 10;
const EXTEND_ICON_SIZE = 9;
const EXPANDED_LINES = 3;

const HBOX_INDEX = 0;
const CONTENTBOX_INDEX = 1;

const MESSAGE_DISPLAY_THRESHOLD = 2;

const Gettext = imports.gettext;
const _ = Gettext.domain('notification-center').gettext;

var MessageListGroup = GObject.registerClass(
class MessageListGroup extends MessageList.MessageListSection {
    _init(source) {
        super._init();
        this._titleBox = new St.BoxLayout({ style_class: 'notification-center-noti-title-box' });
        this._icon = source.createIcon(ICON_SIZE);
        this._titleLabel = new St.Label({ style_class: 'notification-center-noti-title-label',
            y_align: Clutter.ActorAlign.CENTER });
        this._titleLabel.set_text(source.title);
        if (this._icon !== null)
            this._titleBox.add_child(this._icon);
        this._titleBox.add_child(this._titleLabel);
        this.add_child(this._titleBox);
        this.set_child_at_index(this._titleBox, 0);
        this.add_style_class_name('notification-center-message-list-section');

        this._createExtendButton();
        this.add_child(this._extendButton);
        this._extend = false;
        this.updateMessageVisibility();
    }

    _createExtendButton() {
        this._extendButtonBox = new St.BoxLayout({
            x_expand: true,
            y_expand: true,
            vertical: false,
            x_align: Clutter.ActorAlign.CENTER,
            y_align: Clutter.ActorAlign.CENTER,
        });
        this._upIcon = Gio.icon_new_for_string(`${Me.path}/img/tos_desktop_ic_alarm_arrow_up_dark.svg`);
        this._downIcon = Gio.icon_new_for_string(`${Me.path}/img/tos_desktop_ic_alarm_arrow_down_dark.svg`);
        this._extendIcon = new St.Icon({ gicon: this._downIcon,
            style_class: 'notification-center-noti-extend-icon',
            icon_size: EXTEND_ICON_SIZE });
        this._extendNumLabel = new St.Label({ style_class: 'notification-center-noti-extend-label-number' });
        this._extendLabel = new St.Label({ style_class: 'notification-center-noti-extend-label',
            text: _('개 알림 더보기') });
        this._extendButtonBox.add_child(this._extendIcon);
        this._extendButtonBox.add_child(this._extendNumLabel);
        this._extendButtonBox.add_child(this._extendLabel);

        this._extendButton = new St.Button({ child: this._extendButtonBox,
            style_class: 'notification-center-noti-extend-button' });

        this._extendButton.connect('clicked', this._onExtendButtonClicked.bind(this));
    }

    _onExtendButtonClicked() {
        this._extend = !this._extend;
        this._extendNumLabel.visible = !this._extend;
        if (this._extend) {
            this._extendIcon.set_gicon(this._upIcon);
            this._extendLabel.text = _('간략히 보기');
        } else {
            this._extendIcon.set_gicon(this._downIcon);
            this._extendLabel.text = _('개 알림 더보기');
        }
        this.updateMessageVisibility();
    }

    updateMessageVisibility() {
        if (this._messages.length <= MESSAGE_DISPLAY_THRESHOLD) {
            this._extendButton.visible = false;
            for (let message of this._messages.values())
                message.visible = true;
            return;
        }
        this._extendButton.visible = true;
        for (let idx = 0; idx < this._messages.length; idx++)
            this._messages[idx].visible = idx < MESSAGE_DISPLAY_THRESHOLD || this._extend;
        this._extendNumLabel.text = (this._messages.length - MESSAGE_DISPLAY_THRESHOLD).toString();

    }

    /** @override */
    _sync() {
        super._sync();
        if (this._extendButton)
            this.updateMessageVisibility();
    }

});


// eslint-disable-next-line no-unused-vars
var NotificationSectionByGroup = GObject.registerClass({
    Signals: {
        'noti-added': {},
        'noti-destroyed': {},
    },
}, class NotificationSectionByGroup extends Calendar.NotificationSection {
    _init() {
        super._init();
        this.remove_style_class_name('message-list-section');
        this.notiCount = 0;
        this.remove_actor(this._list);
    }

    /** @override */
    _sourceAdded(tray, source) {
        let obj = {
            destroyId: 0,
            notificationAddedId: 0,
            urgency: null,
            messageList: null,
        };

        obj.destroyId = source.connect('destroy', () => {
            this._onSourceDestroy(source, obj);
        });
        obj.notificationAddedId = source.connect('notification-added',
            this._onNotificationAdded.bind(this));

        obj.messageList = new MessageListGroup(source);
        this.add_actor(obj.messageList);
        this._sources.set(source, obj);
    }

    /** @override */
    _onSourceDestroy(source, obj) {
        this.remove_child(obj.messageList);
        source.disconnect(obj.destroyId);
        source.disconnect(obj.notificationAddedId);

        this._sources.delete(source);
    }

    /** @override */
    _onNotificationAdded(source, notification) {
        let message = new Calendar.NotificationMessage(notification);
        let timeLabel =  new Calendar.TimeLabel(notification.datetime);
        let contentBox = message.child.get_child_at_index(HBOX_INDEX).get_child_at_index(CONTENTBOX_INDEX);
        let messageList = this._sources.get(source).messageList;
        contentBox.add_child(timeLabel);
        message.set_style_class_name('notification-center-message');

        timeLabel.add_style_class_name('message-timestamp');

        let newIcon = null;
        if (notification.gicon) {
            newIcon = new St.Icon({ gicon: notification.gicon,
                icon_size: NOTI_ICON_SIZE,
                x_expand: false,
                y_expand: false });
        }
        message.setIcon(newIcon);
        if (newIcon === null) {
            message._iconBin.visible = true;
            message._iconBin.set_style_class_name('message-icon-bin-empty');
        }

        message.setExpandedLines(EXPANDED_LINES);
        message.expand();

        if (this._sources.get(source).urgency === null)
            this._sources.get(source).urgency = notification.urgency;
        this.notiCount++;

        let updatedId = notification.connect('updated', () => {
            let newTimeLabel =  new Calendar.TimeLabel(notification.datetime);
            newTimeLabel.set_style_class_name('message-timestamp');
            contentBox.replace_child(newTimeLabel, timeLabel);
            messageList.moveMessage(message, 0, this.mapped);
            this._updateSectionIndex(source);
            messageList.updateMessageVisibility();
        });
        let destroyId = notification.connect('destroy', () => {
            notification.disconnect(destroyId);
            notification.disconnect(updatedId);
            this.notiCount--;
            this.emit('noti-destroyed');
        });
        messageList.addMessageAtIndex(message, this.mapped, 0);
        this._updateSectionIndex(source);
        this.emit('noti-added');
    }

    // TODO: track number of sections by priority for performance
    _updateSectionIndex(source) {
        let urgency = this._sources.get(source).urgency;
        let newIdx = 0;
        for (let obj of this._sources.values()) {
            if (obj.urgency > urgency)
                newIdx++;
        }
        this.set_child_at_index(this._sources.get(source).messageList, newIdx);
    }

    isEmpty() {
        for (let obj of this._sources.values()) {
            if (!obj.messageList._empty)
                return false;
        }
        return true;
    }

    isUrgent() {
        for (let obj of this._sources.values()) {
            if (obj.urgency === MessageTray.Urgency.CRITICAL)
                return true;
        }
        return false;
    }

    /** @override */
    canClear() {
        for (let obj of this._sources.values()) {
            if (obj.messageList._canClear)
                return true;
        }
        return false;
    }

    /** @override */
    clear() {
        for (let obj of this._sources.values()) {
            if (obj.urgency < MessageTray.Urgency.CRITICAL)
                obj.messageList.clear();
        }
    }

    get getNotiCount() {
        return this.notiCount;
    }
});
